__author__ = 'Rohin Kumar Y'


# Calculate anisotropic 2pCF
from tpcf import *

# antpcf(dat,datR,bins,parmetric,permetric) returns numpy 2d array DD, RR, DR correl
# poserr(xi,DD) returns (1.0+xi)/np.sqrt(DD)


def atpcf(datfile, binspar, binsper, **kwargs):
    """Main function to calculate anisotropic 2pCF. Takes multiple arguments such as randfile, maskfile, calculation estimator etc. for different geometry, cosmology models
    Usage of the package is given in jupyter notebook "Using correlcalc example-anisotropic.nb" and in `main.py`

    All the methods in correlcalc can be imported using the following command

    `from correlcalc import *`

    We first need to define bins (in $c/H_0$ units) to calculate 2pCF. For e.g. to calculate correlation between 0-180Mpc in steps of 6Mpc, we say

    `bins=np.arange(0.002,0.06,0.002)`

    To calculate anisotropic 2pCF using input data file (both ascii and fits files are supported), use `atpcf` method as follows

    `correl2d, poserr=atpcf('/path/to/datfile.dat',bins, randfile='/path/to/randomfile.dat', permetric='apzdth', parmetric='apdz', weights=True)`


    If random file is not available or not provided, we can generate random catalog by providing the mangle mask file in `.ply` format along with specifying the size of the catalog in multiples of size of data catalog (default 2x size). To do this

    `correl2d, poserr=atpcf('/path/to/datfile.dat', bins, maskfile='/path/to/maskfile.ply', permetric='apzdth', parmetric='apdz', weights=True, randfact=3)`

    This returns `correl2d` and `poserr` `numpy` arrays corresponding to anisotropic Two-point correlation and Poisson error

    ### Keyword Arguments
    The following keyword arguments can be included as needed

    #### Data file (Mandatory)

    Data file of galaxy/quasar redshift survey must be passed as the first argument to both `tpcf` and `atpcf` methods.

    **Supported filetypes**: ascii text files with columns, csv files or fits files are all supported. Most files provided by SDSS Value added catalogs should be directly usable.

     **To contain**: Any type of file provided must at least have columns named **Z** (redshift), **RA** (Right Ascension), **DEC** (Declination). These column names can be in any case.

     If one intends to use `weights=True` option (must to obtain accurate results) the data file must also contain radial weights with column title **radial_weight** or **WEIGHT_SYSTOT**

    #### bins (Mandatory)

    A numpy array with ascending values in $c/H_0$ units must be provided as the second argument to both `tpcf` and `atpcf` methods. In case of `atpcf` it automatically creates 2D bins as `bins2d=(bins,bins)` from provided 1D `bins`

    #### `randfile=` Path to random file (semi-Optional)

    If not provided, `maskfile=` argument must be given `.ply` file.

    **Supported filetypes**: ascii text files with columns, csv files or fits files are all supported. Most files provided by SDSS Value added catalogs should be directly usable.

     **To contain**: Any type of file provided must at least have columns named **Z** (redshift), **RA** (Right Ascension), **DEC** (Declination). These column names can be in any case.

     If one intends to use `weights=True` option (must to obtain accurate results) the data file must also contain radial weights with column title **radial_weight** or **WEIGHT_SYSTOT**

     **To add:** In future support for other column titles for weights will be added. To also add calculation of weights from n(z) and for random catalog generation.


    #### `mask=` Path to mangle polygon file (semi-Optional)

    If not provided, `randfile=` argument must be provided.

    **Supported filetypes**: `.ply` file containing Mangle polygons describing survey geometry in the standard format. Most files provided by SDSS Value added catalogs should be directly usable.

    #### `randfact=` (Optional)

    Size of the random catalog in integer multiples of size of data catalog if random catalog file is not provided. Default value is `2`

    #### `weights=` (Optional)

    It is highly recommended to use weights argument by providing `weights=True` to obtain accurate two-point correlation calculations. This picks up radial weights in the prescribed format (with column title **radial_weight** or **WEIGHT_SYSTOT** ) from the data and random files provided.

    If `weights=False`, by default *+1* will be added for each galaxy/random pair found within the bin instead of adding total weight. For more details on weights and references, see http://www.sdss3.org/dr9/tutorials/lss_galaxy.php

    #### Metrics in parallel and perpendicular directions

    Currently calculates anisotropic 2pCF only in angular co-ordinates. Currently only small $\Delta \theta$ and $z \Delta\theta$ are supported. But results can be converted to any cosmology model of choice (ref: https://arxiv.org/pdf/1312.0003.pdf)

    #### `parmetric=` (Mandatory)

    Metric to calculate distance in the line-of-sight direction. Currently only `'apdz'` is supported

    #### `permetric=` (Mandatory)

    Metric to calculate distance perpendicular to the line-of-sight direction. Currently only `'apzdth'` is supported


     **Customization**

     Formulae for calculation of distances in parallel and perpendicular directions is taken from https://arxiv.org/pdf/1312.0003.pdf. Using the formulae in this paper, $\Delta z$ and $z \Delta \theta$ are computed in the `metrics.pyx` file for the above mentioned. `Cython` is chosen for implementation to obtain faster results in building `BallTree`s  calculating `cdist` and to reduce `query` time.

     One can customize metric definitions as per one's need by editing the `metrics.pyx` file. After changing this compile it using `python metricsetup.py build_ext --inplace`

     **To add:**

     Direct calculation of distances in LOS and perpendicular to the LOS to be added to support standard model Cosmology and other popular models. For now, one needs to manually convert the angular bins to  physical distances to get the approximate results


    #### `cosmology='lcdm'` (Optional)

    Used to calculate co-moving distances from redshifts.

    **Available options**:

    `'lcdm'` (default)- for Lambda CDM model

     `'lc'` - for $R_h=ct$ and linear coasting models

     **To add**: `wcdm` and other popular cosmology models soon

    #### `estimator=` (Optional)

    **Available options**:

    `'dp'` - Davis - Peebles estimator (default - fastest)

    `'ls'`- Landy - Szalay estimator

    `'ph'` - Peebles- Hauser estimator

    `'hew'` - Hewitt estimator

    `'h'` - Hamilton estimator

    For more details on estimator formulae see https://arxiv.org/pdf/1211.6211.pdf

    """
    # Default function arguments
    global binsparv
    global binsperv
    global dat
    global datR
    global Nd
    global Nr
    weightsflag = False
    useones = False
    cosmology = 'lcdm'
    sflag = True
    # geometry='flat'
    parmetric = mu
    permetric = flatdistsq
    randcatfact = 2
    estimator = 'dp'
    binsparv = binspar
    binsperv = binsper**2
    randfile = None
    maskfile = None


    # Options for correl calculation estimators and cosmology models
    mlist = ['dp', 'ls', 'ph', 'hew', 'h']
    clist = ['lcdm', 'lc'] # to add wcdm

    if kwargs is not None:
        for key, value in kwargs.items():
            # print (key, value)
            if key.lower() == 'randfile':
                randfile = value

            elif key.lower() == 'randfact':
                randcatfact = value

            elif key.lower() == 'cosmology':
                if value.lower() in clist:
                    cosmology = value.lower()
                else:
                    print("Incorrect Cosmology provided! Using 'lcdm' as default")

            elif key.lower() == 'parmetric':
                if value.lower() == 'apdz':
                    parmetric = APdz
                    binsparv = binspar
                    sflag = False
                elif value.lower() == 'apzdth':
                    parmetric = APzdth
                    binsparv = binspar
                    sflag = False
                elif value.lower() == 'sflat':
                    parmetric = flatdistsq
                    binsparv = binspar**2
                    # sflag = True
                elif value.lower() == 'sopen':
                    parmetric = opendistsq
                    binsparv = binspar**2
                    # sflag = True
                elif value.lower() == 'sclose':
                    parmetric = closedistsq
                    binsparv = binspar**2
                    # sflag = True
                elif value.lower() == 'mu':
                    parmetric = mu
                    binsparv = binspar
                    # sflag = True
                elif value.lower() == 'sparf':
                    parmetric = sparfsq
                    binsparv = binspar**2
                elif value.lower() == 'sparo':
                    parmetric = sparosq
                    binsparv = binspar**2
                elif value.lower() == 'sparc':
                    parmetric = sparcsq
                    binsparv = binspar**2
                else:
                    print("Incorrect parallel metric argument provided!")
            elif key.lower() == 'permetric':
                if value.lower() == 'apzdth':
                    permetric = APzdth
                    binsperv = binsper
                elif value.lower() == 'apdz':
                    permetric = APdz
                    binsperv = binsper
                elif value.lower() == 'sflat':
                    permetric = flatdistsq
                    binsperv = binsper**2
                elif value.lower() == 'sopen':
                    permetric = opendistsq
                    binsperv = binsper**2
                elif value.lower() == 'sclose':
                    permetric = closedistsq
                    binsperv = binsper**2
                elif value.lower() == 'mu':
                    permetric = mu
                    binsperv = binsper
                elif value.lower() == 'sperf':
                    permetric = sperfsq
                    binsperv = binsper**2
                elif value.lower() == 'spero':
                    permetric = sperosq
                    binsperv = binsper**2
                elif value.lower() == 'sperc':
                    permetric = spercsq
                    binsperv = binsper**2
                else:
                    print("Incorrect perpendicular metric provided!")

            elif key.lower() == 'estimator':
                if value.lower() in mlist:
                    estimator = value.lower()
                else:
                    print("Incorrect estimator provided! Using 'dp' as default")

            elif key.lower() == 'mask':
                maskfile = value
            elif key.lower() == 'weights':
                if value is True:
                    weightsflag = True
                elif isinstance(value, str):
                    if value.lower() == 'eq':
                        weightsflag = True
                        useones = True
                else:
                    weightsflag = False
            else:
                print ("key argument `%s` not valid" % key)
    else:
        print ("Refer documentation to enter valid keyword arguments")

    print("Calculating Anisotropic Correlation function with the following parameters")
    print("data file=")
    print(datfile)
    print("random file=")
    print(randfile)
    print("Random catalog size factor(if random file is None)=")
    print(randcatfact)
    print("mask/window file=")
    print(maskfile)
    print ("Cosmology=")
    print(cosmology)
    print("Weights=")
    print(weightsflag)
    print ("Using ones as weights?=")
    print (useones)
    print("perpendicular metric=")
    print(permetric)
    print("parallel metric=")
    print(parmetric)
    print("Correl estimator=")
    print(estimator)
    print ("binsparv=")
    print (binsparv)
    print ("binsperv=")
    print (binsperv)
    print("-----------------------------------")

    if sflag is False:
        # Prepare dat from data file
        dat, weights = datprepz(datfile, 'data', cosmology)
        Nd = len(dat)
        # Prepare datR from random file or generate a random catalog
        if randfile is None:
            randcatsize = randcatfact*Nd
            if maskfile is None:
                print ("Mask file compulsory. Please provide mask='maskfilepath.ply'")
            else:
                datR, rweights = randcatprepz(datfile, randcatsize, maskfile, cosmology)
        else:
            datR, rweights = datprepz(randfile, 'random', cosmology)

    else:
        # Prepare dat from data file
        dat, weights = datprep(datfile, 'data', cosmology)

        Nd = len(dat)
        # Prepare datR from random file or generate a random catalog
        if randfile is None:
            randcatsize = randcatfact*Nd
            if maskfile is None:
                print ("Mask file compulsory. Please provide mask='maskfilepath.ply'")
            else:
                datR, rweights = randcatprep(datfile, randcatsize, maskfile, cosmology)
        else:
            datR, rweights = datprep(randfile, 'random', cosmology)

    Nr = len(datR)
    fact = Nr/Nd
    global adbt
    global arbt

    print ("Creating BallTree for data points using permetric...")
    adbt = BallTree(dat, metric='pyfunc', func=permetric)

    print ("Creating BallTree for random points using permetric...")
    arbt = BallTree(datR, metric='pyfunc', func=permetric)

    rng = np.array([[min(binsparv), max(binsparv)], [min(binsperv), max(binsperv)]])
    print ("Calculating anisotropic 2pCF...")

    # Reference: arXiv: 1211.6211
    if estimator == 'dp':
        if weightsflag is False: # or len(weights) != Nd
            # print (weightsflag)
            # print(len(weights))
            # print(len(datR))
            DD = aDDcalc(dat, binsparv, binsperv, parmetric, permetric, rng)
            DR = aDRcalc(dat, datR, binsparv, binsperv, parmetric, permetric, rng)
        else:
            # if len(rweights)!=len(datR):
            # DD = aDDwcalc(dat, binsq, parmetric, permetric, rng, weights)
            if useones is True or len(weights) != Nd:
                weights = np.ones(Nd)
                rweights = np.ones(Nr)
            print ("Calculating anisotropic DD with weights (parallelized)...\n DD=")
            DD = amulti_autocp(dat, binsparv, binsperv, parmetric, permetric, rng, weights, Nd, pcpus)
            # DR = aRDwcalc(dat, datR, binsq, parmetric, permetric, rng, weights)
            print ("Calculating anisotropic DR with weights (parallelized)...\n DR=")
            DR = amulti_crosscp(dat, datR, binsparv, binsperv, parmetric, permetric, rng, weights, Nr, pcpus)
            # else:
            #     DD=aDDwcalc(dat,binsq,parmetric,permetric,rng,weights)
            #     DR=aDRwcalc(dat,datR,binsq,parmetric,permetric,rng,weights,rweights)

        print ("Using Davis-Peebles estimator")
        correl = fact*(DD/DR)-1.0

    elif estimator == 'ph':
        if weightsflag is False: # or len(weights) != Nd or len(rweights) != len(datR):
            DD = aDDcalc(dat, binsparv, binsperv, parmetric, permetric, rng)
            RR = aRRcalc(datR, binsparv, binsperv, parmetric, permetric, rng)
        else:
            if useones is True or len(weights) != Nd:
                weights = np.ones(Nd)
                rweights = np.ones(Nr)
            print ("Calculating anisotropic DD with weights (parallelized)...\n DD=")
            # DD = aDDwcalc(dat, binsq, parmetric, permetric, rng, weights)
            DD = amulti_autocp(dat, binsparv, binsperv, parmetric, permetric, rng, weights, Nd, pcpus)
            # if len(rweights) != Nr:
            #     RR = aRRcalc(datR, binsparv, binsperv, parmetric, permetric, rng)
            # else:
            print ("Calculating anisotropic RR with weights (parallelized)...\n RR=")
            RR = amulti_autocpr(datR, binsparv, binsperv, parmetric, permetric, rng, rweights, Nr, pcpus)
        print ("Using Peebles-Hauser estimator")
        correl = fact**2*(DD/RR)-1.0
    else:
        if weightsflag is False: # or len(weights) != Nd or len(rweights) != len(datR):
            DD = aDDcalc(dat, binsparv, binsperv, parmetric, permetric, rng)
            RR = aRRcalc(datR, binsparv, binsperv, parmetric, permetric, rng)
            DR = aDRcalc(dat, binsparv, binsperv, parmetric, permetric, rng)
        else:
            if useones is True or len(weights) != Nd:
                weights = np.ones(Nd)
                rweights = np.ones(Nr)
            print ("Calculating anisotropic DD with weights (parallelized)...\n DD=")
            # DD = aDDwcalc(dat, binsq, parmetric, permetric, rng, weights)
            DD = amulti_autocp(dat, binsparv, binsperv, parmetric, permetric, rng, weights, Nd, pcpus)
            # print ("Calculating anisotropic RR with weights (parallelized)...\n RR=")
            # RR = aRRwcalc(datR, binsq, parmetric, permetric, rng, rweights)
            # RR = amulti_autocpr(datR, binsq, parmetric, permetric, rng, rweights, Nr, pcpus)
            # DR = aRDwcalc(dat, datR, binsq, parmetric, permetric, rng, weights)
            print ("Calculating anisotropic DR with weights (parallelized)...\n DR=")
            DR = amulti_crosscp(dat, datR, binsparv, binsperv, parmetric, permetric, rng, weights, Nr, pcpus)
            # if len(rweights) != Nr:
            #     RR = aRRcalc(datR, binsparv, binsperv, parmetric, permetric, rng)
            # else:
            print ("Calculating anisotropic RR with weights (parallelized)...\n RR=")
            RR = amulti_autocpr(datR, binsparv, binsperv, parmetric, permetric, rng, rweights, Nr, pcpus)
        if estimator == 'ls':
            print ("Using Landy-Szalay estimator")
            correl = fact**2*(DD/RR)-2.0*fact*(DR/RR)+1.0
        elif estimator == 'hew':
            print ("Using Hewett estimator")
            correl = fact**2*(DD/RR)-fact*(DR/RR)
        elif estimator == 'h':
            print ("Using Hamilton estimator")
            correl = (DD*RR)/DR**2 - 1.0
    correlerr = poserr(correl, DD)
    print("Anisotropic Two-point correlation=")
    print (correl, correlerr)
    return correl, correlerr


def aDDcalc(dat, binspar, binsper, parmetric, permetric, rng):
    print ("Calculating anisotropic DD...\n DD=")
    dd = np.zeros((len(binspar)-1, len(binsper)-1))
    for i in tqdm(range(len(dat))):
        ind = adbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], dat[j[j>i]], parmetric)[0]
            # print("dist0")
            # print dist0
            dist1 = dist.cdist([dat[i], ], dat[j[j>i]], permetric)[0]
            # print("dist1")
            # print dist1
            # print np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper))[0]
            dd += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper))[0]
            # print ("rng")
            # print rng
            # print("binspar")
            # print binspar
            # print("binsper")
            # print binsper
            # print dd
    dd[dd == 0] = 1.0
    # Nd = len(dat)
    # DD = dd/(Nd*(Nd-1.0))
    print (dd)
    return dd


def aRRcalc(datR, binspar, binsper, parmetric, permetric, rng):
    print ("Calculating anisotropic RR...\n RR=")
    rr = np.zeros((len(binspar)-1, len(binsper)-1))
    # rrbt = BallTree(datR, metric='pyfunc', func=permetric)
    for i in tqdm(range(len(datR))):
        ind = arbt.query_radius(datR[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([datR[i], ], datR[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([datR[i], ], datR[j[j>i]], permetric)[0]
            rr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper))[0]
    rr[rr == 0] = 1.0
    # Nr = len(datR)
    # RR = rr/(Nr*(Nr-1.0))
    print (rr)
    return rr


def aDRcalc(dat, datR, binspar, binsper, parmetric, permetric, rng):
    print ("Calculating anisotropic DR...\n DR=")
    dr = np.zeros((len(binspar)-1, len(binsper)-1))
    # rrbt = BallTree(datR, metric='pyfunc', func=permetric)
    for i in tqdm(range(len(dat))):
        ind = arbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], datR[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([dat[i], ], datR[j[j>i]], permetric)[0]
            dr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper))[0]
    dr[dr == 0] = 1.0
    # Nd = len(dat)
    # Nr = len(datR)
    # DR = dr/(Nd*Nr)
    print (dr)
    return dr


def aDDwcalc(dat, binspar, binsper, parmetric, permetric, rng, weights):
    print ("Calculating anisotropic DD with weights...\n DD=")
    dd = np.zeros((len(binspar)-1, len(binsper)-1))
    # ddbt = BallTree(dat, metric='pyfunc', func=permetric)
    for i in tqdm(range(len(dat))):
        ind = adbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], dat[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([dat[i], ], dat[j[j>i]], permetric)[0]
            dd += np.histogram2d(dist0, dist1, range=rng, bins=(bins, bins), weights=weights[j[j>i]])[0]
    dd[dd == 0] = 1.0
    # Nd = len(dat)
    # DD = dd/(Nd*(Nd-1.0)) # factor of 2 cancels with 1/2 that needs to be done to remove double counting of pairs
    # print (dd)
    return dd


def aRRwcalc(datR, binspar, binsper, parmetric, permetric, rng, rweights):
    print ("Calculating anisotropic RR with weights...\n RR=")
    rr = np.zeros((len(binspar)-1, len(binsper)-1))
    for i in tqdm(range(len(datR))):
        ind = arbt.query_radius(datR[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([datR[i], ], datR[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([datR[i], ], datR[j[j>i]], permetric)[0]
            rr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=rweights[j[j>i]])[0]
    rr[rr == 0] = 1.0
    # Nr = len(datR)
    # RR = rr/(Nr*(Nr-1.0)) # factor of 2 cancels with 1/2 that needs to be done to remove double counting of pairs
    # print (rr)
    return rr


def aDRwcalc(dat, datR, binspar, binsper, parmetric, permetric, rng, weights, rweights):
    print ("Calculating anisotropic DR with weights...\n DR=")
    dr = np.zeros((len(binspar)-1, len(binsper)-1))
    # rrbt = BallTree(datR, metric='pyfunc', func=permetric)
    for i in tqdm(range(len(dat))):
        ind = arbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], datR[j], parmetric)[0]
            dist1 = dist.cdist([dat[i], ], datR[j], permetric)[0]
            dr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=rweights[j])[0]
    dr[dr == 0] = 1.0
    # Nd = len(dat)
    # Nr = len(datR)
    # DR = dr/(Nd*Nr)
    # print (dr/2.0)
    return dr/2.0


def aRDwcalc(dat, datR, binspar, binsper, parmetric, permetric, rng, weights):
    print ("Calculating anisotropic RD with weights...\n DR=")
    dr = np.zeros((len(binspar)-1, len(binsper)-1))
    # bt = BallTree(dat, metric='pyfunc', func=permetric)
    for i in tqdm(range(len(datR))):
        ind = arbt.query_radius(datR[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([datR[i], ], dat[j], parmetric)[0]
            dist1 = dist.cdist([datR[i], ], dat[j], permetric)[0]
            dr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=weights[j])[0]
    dr[dr == 0] = 1.0
    # DR = dr/(Nd*Nr)
    # print (dr/2.0)
    return dr/2.0

def aDDwcalcp(dat, binspar, binsper, parmetric, permetric, rng, weights, rNd, multi=False, queue=0):
    dd = np.zeros((len(binspar)-1, len(binsper)-1))
    # ddbt = BallTree(dat, metric='pyfunc', func=permetric)
    for i in tqdm(rNd):
        ind = adbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], dat[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([dat[i], ], dat[j[j>i]], permetric)[0]
            dd += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=weights[j[j>i]])[0]
    if multi:
        queue.put(dd)
    else:
        return dd
    # print (DD)
    return dd


def aRRwcalcp(datR, binspar, binsper, parmetric, permetric, rng, rweights, rNr, multi=False, queue=0):
    rr = np.zeros((len(binspar)-1, len(binsper)-1))
    # rrbt = BallTree(datR, metric='pyfunc', func=permetric)
    for i in tqdm(rNr):
        ind = arbt.query_radius(datR[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([datR[i], ], datR[j[j>i]], parmetric)[0]
            dist1 = dist.cdist([datR[i], ], datR[j[j>i]], permetric)[0]
            rr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=rweights[j[j>i]])[0]
    if multi:
        queue.put(rr)
    else:
        return rr
    # rr[rr == 0] = 1.0
    # Nr = len(datR)
    # RR = rr/(Nr*(Nr-1.0)) # factor of 2 cancels with 1/2 that needs to be done to remove double counting of pairs
    # print (RR)
    return rr


def aDRwcalcp(dat, datR, binspar, binsper, parmetric, permetric, rng, rweights, rNd, multi=False, queue=0):
    # print ("Calculating anisotropic DR with weights (parallelized)...\n DR=")
    dr = np.zeros((len(binspar)-1, len(binsper)-1))
    # rrbt = BallTree(datR, metric='pyfunc', func=permetric)
    for i in tqdm(rNd):
        ind = arbt.query_radius(dat[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([dat[i], ], datR[j], parmetric)[0]
            dist1 = dist.cdist([dat[i], ], datR[j], permetric)[0]
            dr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=rweights[j])[0]
    if multi:
        queue.put(dr)
    else:
        return dr
    # print (DR)
    return dr


def aRDwcalcp(dat, datR, binspar, binsper, parmetric, permetric, rng, weights, rNr, multi=False, queue=0):
    # print ("Calculating anisotropic RD with weights (parallelized)...\n DR=")
    dr = np.zeros((len(binspar)-1, len(binsper)-1))
    # bt = BallTree(dat, metric='pyfunc', func=permetric)
    for i in tqdm(rNr):
        ind = adbt.query_radius(datR[i].reshape(1, -1), max(binsper))
        for j in ind:
            dist0 = dist.cdist([datR[i], ], dat[j], parmetric)[0]
            dist1 = dist.cdist([datR[i], ], dat[j], permetric)[0]
            dr += np.histogram2d(dist0, dist1, range=rng, bins=(binspar, binsper), weights=weights[j])[0]
    if multi:
        queue.put(dr)
    else:
        return dr
    return dr


def amulti_autocp(dat, binspar, binsper, parmetric, permetric, rng, weights, Nd, CORES=pcpus):

    DD = np.zeros((len(binspar)-1, len(binsper)-1))
    queues = [RetryQueue() for i in range(CORES)]
    args = [(dat, binspar, binsper, parmetric, permetric, rng, weights, range(int(Nd*i/CORES),int(Nd*(i+1)/CORES)), True, queues[i]) for i in range(CORES)]
    jobs = [Process(target=aDDwcalcp, args=(a)) for a in args]
    for j in jobs: j.start()
    for q in queues: DD += q.get()
    for j in jobs: j.join()
    DD[DD == 0] = 1.0
    # DD = DD/(Nd*(Nd-1.0)) # factor of 2 cancels with 1/2 that needs to be done to remove double counting of pairs
    print (DD)
    return DD


def amulti_autocpr(datR, binspar, binsper, parmetric, permetric, rng, rweights, Nr, CORES=pcpus):

    RR = np.zeros((len(binspar)-1, len(binsper)-1))
    queues = [RetryQueue() for i in range(CORES)]
    args = [(datR, binspar, binsper, parmetric, permetric, rng, rweights, range(int(Nr*i/CORES),int(Nr*(i+1)/CORES)), True, queues[i]) for i in range(CORES)]
    jobs = [Process(target=aRRwcalcp, args=(a)) for a in args]
    for j in jobs: j.start()
    for q in queues: RR += q.get()
    for j in jobs: j.join()
    RR[RR == 0] = 1.0
    # RR = RR/(Nr*(Nr-1.0)) # factor of 2 cancels with 1/2 that needs to be done to remove double counting of pairs
    print (RR)
    return RR


def amulti_crosscp(dat, datR, binspar, binsper, parmetric, permetric, rng, weights, Nr, CORES=pcpus):

    DR = np.zeros((len(binspar)-1, len(binsper)-1))
    queues = [RetryQueue() for i in range(CORES)]
    args = [(dat, datR, binspar, binsper, parmetric, permetric, rng, weights, range(int(Nr*i/CORES),int(Nr*(i+1)/CORES)), True, queues[i]) for i in range(CORES)]
    jobs = [Process(target=aRDwcalcp, args=(a)) for a in args]
    for j in jobs: j.start()
    for q in queues: DR += q.get()
    for j in jobs: j.join()
    DR[DR == 0] = 1.0
    # Nd=len(dat)
    # DR = DR/(Nd*Nr)
    print (DR/2.0)
    return DR/2.0



__author__ = 'Rohin Kumar Y'
# from fileios import *
# msg = 'Enter Absolute Path to file: '
# f_name = raw_input(msg).strip()
# Ran multiple tests for most methods in the package... need to clean up
# path = file_data_and_path(f_name)
# if path != None:
#       print 'Path:',path
# import tpcf
# from fileios import *
# from comovdist import *
from datprep import *
from tpcf import *
import matplotlib.pyplot as plt
from genrand import *
from runtimeupdate import *
# from metrics import *
# t1 = checkdatfile('./testfile.dat')
# print t1
#
# t2=inputfiles('data')
# print t2
# %matplotlib osx
from datvis import *

# dat=checkinfile('./testfile.dat','data')
# print dat
from metrics import *
from tpcf import *
# dat=readinfile('./testfile.dat','data')
# dat=datprep('./testfile.dat','data','lcdm')
from antpcf import *
# bins=np.arange(0.002,0.082,0.002)
# tpcf('./testfile.dat',bins,randfile='./testfile.dat',method='ls')
# tpcf('./testfile.dat',bins,mask='/Users/rohin/Documents/ipy_notebooks/galsurveystudy/masks/boss_geometry_2011_06_10.ply',cosmology='lc',method='ls')
# bins=np.arange(0,0.201,0.01)
# tpcf(dat,dat,3,bins,flatdistsq,'ls')
# antpcf(dat,dat,bins,flatdistsq,flatdistsq,'ls')
# z=np.array(dat['Z'])
# ra=np.array(dat['ra'])
# dec=np.array(dat['Dec'])
# hpix=healpixmap(ra,dec)
# print(hpix)
# hu.mollview(hpix,rot=180)
# zr=randz(z,2*len(dat['Z']))
# print(zr)
# spinner.text='Generating Random RA, DEC...'
# spinner.start()
# ra,dec=randang('/Users/rohin/Downloads/mask.dr72safe0.ply',2*len(dat['Z']))
# spinner.stop()
# print(ra,dec)
# plt.hist(z, 10, normed=True, alpha=0.5, label='hist')
# x_grid=np.linspace(min(z),max(z),2*len(dat['Z']))#check  nb file
# kdepdf=kde(z,x_grid,bandwidth=1e-3)#check  nb file#,bandwidth=0.01
# plt.plot(x_grid, kdepdf, color='r', alpha=0.5, lw=3, label='kde')
# plt.show()
# bins=np.arange(0,0.1,0.01)
# print(autocorr(dat,bins,flatdistsq))
# print(crosscorr(dat,dat,bins,flatdistsq))
# print(closedistsq(dat[1],dat[2]))
# mask=readmaskfile('/Users/rohin/Downloads/mask.dr72safe0.ply')
#
# z=dat['Z']
# print(z)
# np.vectorize(Ez)
# ez=Ez(z)
# print ez
# np.vectorize(DC_LCDM)
# s=DC_LCDM(z)
# s=comov(z,'lcdm')
# print s
# For 2pCF for LCDM model
from correlcalc import *
bins = np.arange(0.002, 0.082, 0.002)
correldr72 = tpcf('/Users/rohin/Downloads/DR7-Full.ascii', bins, mask='/Users/rohin/Documents/ipy_notebooks/galsurveystudy/masks/window.dr72safe0.ply', randfact=2)
import matplotlib.pyplot as plt
binMpc = 3000*bins
plt.plot(binMpc[1:], correldr72[0], 'ro-')
plt.show()
plt.yscale('log')
plt.plot(binMpc[4:], correldr72[0][3:], 'ro-')
plt.show()
plt.yscale('log')
plt.xscale('log')
plt.plot(binMpc[3:], correldr72[0][2:], 'ro-')
plt.show()
# For anisotropic 2pCF using delta Z and Z delta theta as in arXiv: 1312.0003
from antpcf import *
bins=np.arange(0.01, 0.201, 0.01)
atpcf('./testfile.dat', bins, randfile='./testfile.dat', estimator='ls', permetric='apzdth', parmetric='apdz')



from libc.math cimport sin, cos, sqrt, acos, sinh, cosh


def flatdistsq (double[:] x, double[:] y):
    cdef double s1 = x[0]
    cdef double s2 = y[0]
    return s1**2+s2**2-2.0*s1*s2*mu(x,y)


def opendistsq (double[:] x, double[:] y):
    cdef double res = 0.0
    cdef double K = -1.0
    cdef double s1 = sinh(x[0]*0.007)
    cdef double s2 = sinh(y[0]*0.007)
    cdef double c1 = cosh(x[0]*0.007)
    cdef double c2 = cosh(y[0]*0.007)
    cdef double costheta = mu(x,y)
    res = s1*s1+s2*s2-2.0*s1*s2*c1*c2*costheta-K*s1*s1*s2*s2*sqrt(1.0+costheta*costheta)
    return res*20408.1633


def closedistsq (double[:] x, double[:] y):
    cdef double res = 0.0
    cdef double K = 1.0
    cdef double s1 = sin(x[0]*0.007)
    cdef double s2 = sin(y[0]*0.007)
    cdef double c1 = cos(x[0]*0.007)
    cdef double c2 = cos(y[0]*0.007)
    cdef double costheta = mu(x,y)
    res = s1*s1+s2*s2-2.0*s1*s2*c1*c2*costheta-K*s1*s1*s2*s2*sqrt(1.0+costheta*costheta)
    return res*20408.1633


def APzdth(double[:] x, double[:] y):
    return (x[0]+y[0])*0.5*acos(mu(x,y))


def APdz(double[:] x, double[:] y):
    return abs(x[0]-y[0])


def mu(double[:] x, double[:] y):
    return sin(x[2])*sin(y[2])+cos(x[2])*cos(y[2])*cos(x[1]-y[1])


def sparfsq(double[:] x, double[:] y):
    return flatdistsq(x, y)*(mu(x,y))**2


def sperfsq(double[:] x, double[:] y):
    return flatdistsq(x, y)*(1.0-(mu(x,y))**2)


def sparosq(double[:] x, double[:] y):
    return opendistsq(x, y)*(mu(x,y))**2


def sperosq(double[:] x, double[:] y):
    return opendistsq(x, y)*(1.0-(mu(x,y))**2)


def sparcsq(double[:] x, double[:] y):
    return closedistsq(x, y)*(mu(x,y))**2


def spercsq(double[:] x, double[:] y):
    return closedistsq(x, y)*(1.0-(mu(x,y))**2)



__author__ = 'Rohin Kumar Y'
import numpy as np
from fileios import *
from genrand import *
from comovdist import *
from math import pi
# calculate comoving distances and convert to 3xN np array
# comov(z,**kwargs)
# **kwargs 'lcdm', 'lc' model


def datprep(fname, ftype, model):
    """Method to convert input ascii files into 3xN matrices needed for metric distance calculations. Calculates comoving distances as per input model."""
    if fname.lower().endswith('.fits'):
        data = readfitsfile(fname, ftype)
    else:
        data = readinfile(fname, ftype)
    weights = np.array([])
    print ("Reading column data from input file...")
    for x in data.colnames:
        if x.lower() == 'z':
            z = np.array(data[x])
        elif x.lower() == 'ra':
            ra = np.array(data[x])
        elif x.lower() == 'dec':
            dec = np.array(data[x])
        elif x.lower() == 'radial_weight':
            weights = np.array(data[x])
        elif x.lower() == 'weight_fkp':
            weights = np.array(data[x])
        elif x.lower() == 'weight_systot':
            weights = np.array(data[x])
        # elif x.lower()=='nz':
        #     weights=1.0/(1.0+4.0*np.array(data[x]))
        #     weights=weights/np.mean(weights)
            # print (weights)
        else:
            pass
    # print ("Calcuating comoving distances and converting angles to radians...")
    s = comovp(z, model)
    rar = ra*pi/180.0
    decr = dec*pi/180.0
    print ("Preparing %s into 3xN matrices in [s,rar,decr] format..." % ftype)
    dat = np.array([s, rar, decr])
    dat.reshape(3, len(data))
    dat = dat.transpose()
    # print ("Printing 3xN array")
    # print(dat)
    return dat, weights


def datprepz(fname, ftype, model):
    """Method to convert input ascii files into 3xN matrices needed for metric distance calculations. passes z instead of s in matrices."""
    if fname.lower().endswith('.fits'):
        data = readfitsfile(fname, ftype)
    else:
        data = readinfile(fname, ftype)
    weights = np.array([])
    print ("Reading column data from input file...")
    for x in data.colnames:
        if x.lower() == 'z':
            z = np.array(data[x])
        elif x.lower() == 'ra':
            ra = np.array(data[x])
        elif x.lower() == 'dec':
            dec = np.array(data[x])
        elif x.lower() == 'radial_weight':
            weights = np.array(data[x])
        elif x.lower() == 'weight_fkp':
            weights = np.array(data[x])
        elif x.lower() == 'weight_systot':
            weights = np.array(data[x])
        # elif x.lower()=='nz':
        #     weights=1.0/(1.0+4.0*np.array(data[x]))
        #     weights=weights/np.mean(weights)
            # print (weights)
        else:
            pass
    # print ("Converting angles to radians...")
    rar = ra*pi/180.0
    decr = dec*pi/180.0
    print ("Preparing %s into 3xN matrices in [z,rar,decr] format..." % ftype )
    dat = np.array([z, rar, decr])
    dat.reshape(3, len(data))
    dat = dat.transpose()
    # print ("Printing 3xN array")
    # print(dat)
    return dat, weights


def randcatprep(datfname, randcatsize, maskfile, model):
    """Method to generate random catalog from mangle mask and input redshift distribution for given cosmology"""
    print("Generating random catalog of %d size in file randcat.dat... "% randcatsize)
    if datfname.lower().endswith('.fits'):
        data = readfitsfile(datfname, 'internal')
    else:
        data = readinfile(datfname, 'internal')

    for x in data.colnames:
        if x.lower() == 'z':
            z = np.array(data[x])
        # elif x.lower()=='nz':
        #     rweights=1.0/(1.0+4.0*np.array(data[x]))
        #     rweights=weights/np.mean(rweights)
        # else:
        #
    zr, rweights = randz(z, randcatsize)
    ra, dec = randang(maskfile, randcatsize)
    print ("Generated random catalog...")
    print ("zr = ")
    print (zr)
    print ("random ra =")
    print (ra)
    print ("random dec =")
    print (dec)
    print ("radial weights =")
    print (rweights)
    rar = ra*pi/180.0
    decr = dec*pi/180.0
    rcatfname = "randcat.dat"# %(datfname)
    storerandcat(zr, ra, dec, rweights, rcatfname)
    s = comovp(zr, model)
    print ("Preparing random catalog into 3xN matrices in [s,rar,decr] format" )
    datR = np.array([s, rar, decr])
    datR.reshape(3, len(zr))
    datR = datR.transpose()
    # print(datR)
    return datR, rweights


def randcatprepz(datfname, randcatsize, maskfile, model):
    """Method to generate random catalog from mangle mask and input redshift distribution for given cosmology"""
    print("Generating random catalog of %d size in file randcat.dat... "% randcatsize)
    if datfname.lower().endswith('.fits'):
        data = readfitsfile(datfname, 'internal')
    else:
        data = readinfile(datfname, 'internal')

    for x in data.colnames:
        if x.lower() == 'z':
            z = np.array(data[x])
        # elif x.lower()=='nz':
        #     rweights=1.0/(1.0+4.0*np.array(data[x]))
        #     rweights=weights/np.mean(rweights)
        # else:
        #
    zr, rweights = randz(z, randcatsize)
    ra, dec = randang(maskfile, randcatsize)
    rar = ra*pi/180.0
    decr = dec*pi/180.0
    rcatfname = "randcat.dat"# %(datfname)
    storerandcat(zr, ra, dec, rweights, rcatfname)
    print ("Preparing random catalog into 3xN matrices in [zr,rar,decr] format" )
    datR = np.array([zr, rar, decr])
    datR.reshape(3, len(zr))
    datR = datR.transpose()
    # print(datR)
    return datR, rweights




__author__ = 'Rohin Kumar Y'
from scipy import integrate
import numpy as np
import math as m
#from . import *
from param import *
# from correlcalc import *
from multiprocessing import Pool
from multiprocessing import cpu_count
# Om = param.Om
# Ol = param.Ol

pcpus = cpu_count() - 1


def Ezs(zv):
    """Hubble papameter in H0 units"""
    return 1.0/m.sqrt(Om*(1.0+zv)**3+(1.0-Om-Ol)*(1.0+zv)**2+Ol)


Ez = np.vectorize(Ezs)


def DC_LCDMs(z):
    """Method to calculate comoving distance for LCDM model"""
    return integrate.quad(Ez, 0, z)[0]


DC_LCDM = np.vectorize(DC_LCDMs)


def DC_LC(z):
    """Method for comoving distance in Linear Coasting model"""
    return np.log(1.0+z)


def comov(z, model):
    """Method to calculate comoving distance of given redshifts for input model. Units in c/H0"""
    # More models such as wcdm to be added
    print ("Calculating comoving distances...")
    if model == 'lcdm':
        return DC_LCDM(z)
    elif model == 'lc':
        return DC_LC(z)
    else:
        print("Only 'lcdm' and 'lc' models supported for now")
        return None


def comovp(z, model):
    """Method to calculate comoving distance of given redshifts for input model. Units in c/H0"""
    # More models such as wcdm to be added
    print ("Calculating comoving distances (parallelized)...")
    pool = Pool(processes=pcpus)
    if model == 'lcdm':
        res = pool.map(DC_LCDM, z)
        pool.close()
        pool.join()
        return res
    elif model == 'lc':
        res = pool.map(DC_LC, z)
        pool.close()
        pool.join()
        return res
    else:
        print("Only 'lcdm' and 'lc' models supported for now")
        return None



# Set global parameters here related to Cosmology (more to be added)
# Om= matter density Ol=dark energy density at the present epoch
# Planck 2015
# Om = 0.3121
# Ol = 0.6879
# WMAP9
# Om = 0.279
# Ol = 0.721
# Default
Om = 0.3
Ol = 0.7



__author__ = 'Rohin Kumar Y'
from scipy.stats import gaussian_kde
import numpy as np
from fileios import *

# Need to parallelize these... Very slow indeed! :(


def kde(x, x_grid, bandwidth=0.2):
    """Kernel Density Estimation with Scipy"""
    kdev = gaussian_kde(x, bw_method=bandwidth / x.std(ddof=1))
    return kdev.evaluate(x_grid)


def generate_rand_from_pdf(pdf, x_grid, N):
    """Method to generate 'N' no. of random numbers from input probability distribution function (pdf) in form of kernel density"""
    cdf = np.cumsum(pdf)
    cdf = cdf / cdf[-1]
    values = np.random.rand(N)
    value_bins = np.searchsorted(cdf, values)
    random_from_cdf, nz = x_grid[value_bins], cdf[value_bins]
    print (nz)
    return random_from_cdf, nz


def randang(maskfile, randcatsize):
    """Method to calculate RA and DEC from mangle .ply file"""
    mangle = readmaskfile(maskfile)
    rar, decr = mangle.genrand(randcatsize)
    rar = np.asfarray(rar)
    decr = np.asfarray(decr)
    return rar, decr


def randz(z, randcatsize):
    """Method to calculate random redshift values from input redshift distribution"""
    x_grid = np.linspace(min(z), max(z), num=randcatsize)
    pdf = kde(z, x_grid, bandwidth=1e-3)
    randzv, nz = generate_rand_from_pdf(pdf, x_grid, randcatsize)
    # randzv = np.array(randzv, np.double)
    rweights = 1.0/(1.0+40000*nz)
    rweights = rweights/np.mean(rweights)
    return randzv, rweights
